﻿Imports Microsoft.CSharp
Imports System.Reflection
Imports System.IO
Imports System.Text
Imports System.Text.RegularExpressions

Public Delegate Function Parser(ByVal Input As String, ByVal ErrorWriter As TextWriter) As String

''' <summary>
''' Abstract base class for creating Coco/R-based parsers from a specified grammar
''' </summary>
Public MustInherit Class CocoRParserGenerator

    Public Enum Notification
        Initializing
        CompilingGrammar
        CompilingParser
    End Enum

    Public Enum Failure
        InvalidGrammar
        InvalidParser
    End Enum

    Public Class ProgressNotifcationEventArgs : Inherits EventArgs
        Public Notification As Notification
    End Class

    Public Class GenerationFailureEventArgs : Inherits EventArgs
        Public Failure As Failure
        Public Data As Object
    End Class

    Public Class GenerationCompletedEventArgs : Inherits EventArgs
        Public Parser As Parser
    End Class

    Public Event GenerationFailure As EventHandler(Of GenerationFailureEventArgs)
    Public Event ProgessNotification As EventHandler(Of ProgressNotifcationEventArgs)
    Public Event GenerationCompleted As EventHandler(Of GenerationCompletedEventArgs)

    Protected Overridable Sub OnGenerationFailure(ByVal e As GenerationFailureEventArgs)
        RaiseEvent GenerationFailure(Me, e)
    End Sub

    Protected Overridable Sub OnProgressNotification(ByVal e As ProgressNotifcationEventArgs)
        RaiseEvent ProgessNotification(Me, e)
    End Sub

    Protected Overridable Sub OnGenerationCompleted(ByVal e As GenerationCompletedEventArgs)
        RaiseEvent GenerationCompleted(Me, e)
    End Sub

    Private ReadOnly m_Executable As String
    Private ReadOnly m_WorkingDirectory As String
    Private ReadOnly m_FrameDirectory As String
    Private ReadOnly m_Extension As String
    Private ReadOnly m_ProviderGenerator As Func(Of CodeDom.Compiler.CodeDomProvider)

    Private ReadOnly ErrRegex As New Regex("^\s*--(.+?)\s*$", RegexOptions.Compiled Or RegexOptions.Multiline)
    Private ReadOnly NumErrRegex As New Regex("(\d+) errors detected\s*$", RegexOptions.Compiled)

    Private m_Working As Boolean = False
    Private m_AsyncCall As Action(Of String)

    ''' <summary>
    ''' Creates a new Coco/R parser generator
    ''' </summary>
    ''' <param name="CocoExecutable">Path to coco.exe</param>
    ''' <param name="WorkingDirectory">Working directory</param>
    ''' <param name="FrameDirectory">Directory containing frame files</param>
    ''' <param name="CreatedExtension">The extension of files that coco.exe will generate</param>
    ''' <param name="CodeProvider">Constructor for a code generation object</param>
    ''' <remarks>All pathes are relative to <paramref>WorkingDirectory</paramref></remarks>
    Public Sub New(ByVal CocoExecutable As String, _
                   ByVal WorkingDirectory As String, _
                   ByVal FrameDirectory As String, _
                   ByVal CreatedExtension As String, _
                   ByVal CodeProvider As Func(Of CodeDom.Compiler.CodeDomProvider))

        m_Executable = CocoExecutable
        m_WorkingDirectory = WorkingDirectory
        m_FrameDirectory = FrameDirectory
        m_Extension = CreatedExtension
        m_ProviderGenerator = CodeProvider
    End Sub

    ''' <summary>
    ''' Starts asynchronous generation of grammar 
    ''' </summary>
    ''' <param name="Grammar">The Coco/R-grammar to be compiled</param>
    Public Sub GenerateAsync(ByVal Grammar As String)
        If m_Working Then Throw New InvalidOperationException("Another generation process is currently running")
        m_AsyncCall = AddressOf RunAsync
        m_Working = True
        m_AsyncCall.BeginInvoke(Grammar, AddressOf Callback, Nothing)
    End Sub

    Private Sub Callback(ByVal ar As IAsyncResult)
        m_AsyncCall.EndInvoke(ar)
        m_Working = False
    End Sub

    Private Sub RunAsync(ByVal Grammar As String)
        Dim RelOutputDirectory = ""
        Dim OutputDirectory = ""

        OnProgressNotification( _
               New ProgressNotifcationEventArgs With {.Notification = Notification.Initializing})

        For i = 1 To Integer.MaxValue
            RelOutputDirectory = String.Format("${0}", i)
            OutputDirectory = Path.Combine(m_WorkingDirectory, RelOutputDirectory)

            If Not Directory.Exists(OutputDirectory) Then
                Directory.CreateDirectory(OutputDirectory)
                File.SetAttributes(OutputDirectory, FileAttributes.Hidden)
                Exit For
            End If
        Next

        Dim GrammarFileRel = "temp.atg"
        Dim GrammarFilename = Path.Combine(OutputDirectory, GrammarFileRel)
        My.Computer.FileSystem.WriteAllText(GrammarFilename, Grammar, False)

        OnProgressNotification( _
               New ProgressNotifcationEventArgs With {.Notification = Notification.CompilingGrammar})

        Dim Coco = Process.Start(New ProcessStartInfo() With { _
                                 .FileName = Path.Combine(m_WorkingDirectory, m_Executable), _
                                 .Arguments = Path.Combine(RelOutputDirectory, GrammarFileRel) & " -frames " & m_FrameDirectory & " -o " & RelOutputDirectory, _
                                 .UseShellExecute = False, _
                                 .WorkingDirectory = m_WorkingDirectory, _
                                 .CreateNoWindow = True, _
                                 .RedirectStandardOutput = True})

        Dim Output = Coco.StandardOutput.ReadToEnd

        Dim NumErrs = Integer.Parse(NumErrRegex.Match(Output).Groups(1).Value)

        If NumErrs <> 0 Then
            Cleanup(OutputDirectory)

            Dim Errs = From m In ErrRegex.Matches(Output).Cast(Of Match)() _
                       Let ErrLine = m.Groups(1).Value _
                       Select ErrLine.Trim

            OnGenerationFailure( _
                   New GenerationFailureEventArgs With { _
                                    .Failure = Failure.InvalidGrammar, _
                                    .Data = Errs.ToList()})
            Exit Sub
        End If

        OnProgressNotification( _
               New ProgressNotifcationEventArgs With {.Notification = Notification.CompilingParser})

        Dim ExtendFilename = Function(s As String) String.Format("{0}.{1}", s, m_Extension)

        Using CodeProvider = m_ProviderGenerator()
            Dim CompilerParameters = New System.CodeDom.Compiler.CompilerParameters() With {.GenerateInMemory = True}

            Dim AbsFrameDirectory = Path.Combine(m_WorkingDirectory, m_FrameDirectory)

            Dim Res = CodeProvider.CompileAssemblyFromFile(CompilerParameters, _
                                                           Path.Combine(AbsFrameDirectory, ExtendFilename("Main")), _
                                                           Path.Combine(OutputDirectory, ExtendFilename("Scanner")), _
                                                           Path.Combine(OutputDirectory, ExtendFilename("Parser")))

            If Res.Errors.Count = 0 Then
                Dim Asm = Res.CompiledAssembly
                Dim MainClass = Asm.GetType("Main")
                Dim Result = Function(data As String, err As TextWriter) MainClass.GetMethod("Parse").Invoke(Nothing, New Object() {data, err}).ToString()
                Cleanup(OutputDirectory)
                OnGenerationCompleted( _
                       New GenerationCompletedEventArgs With { _
                           .Parser = Result})
            Else
                Dim Errs As New List(Of String)
                For i = 0 To Res.Errors.Count - 1
                    Errs.Add(Res.Errors(i).ErrorText)
                Next
                Cleanup(OutputDirectory)
                OnGenerationFailure( _
                       New GenerationFailureEventArgs With { _
                                        .Failure = Failure.InvalidParser, _
                                        .Data = Errs})
            End If
        End Using
    End Sub

    Private Sub Cleanup(ByVal WorkingDir As String)
        For Each Filename In Directory.GetFiles(WorkingDir)
            File.Delete(Filename)
        Next
        Directory.Delete(WorkingDir)
    End Sub

End Class

Public Class CSharpParserGenerator : Inherits CocoRParserGenerator
    Public Sub New(ByVal CocoExecutable As String, _
                   ByVal WorkingDirectory As String, _
                   ByVal FrameDirectory As String)

        MyBase.New(CocoExecutable, WorkingDirectory, FrameDirectory, "cs", Function() New Microsoft.CSharp.CSharpCodeProvider)
    End Sub
End Class

Public Class VBParserGenerator : Inherits CocoRParserGenerator
    Public Sub New(ByVal CocoExecutable As String, _
                   ByVal WorkingDirectory As String, _
                   ByVal FrameDirectory As String)

        MyBase.New(CocoExecutable, WorkingDirectory, FrameDirectory, "vb", Function() New Microsoft.VisualBasic.VBCodeProvider)
    End Sub
End Class